#include "genesis.h"
#include "z80.h"
#include "video.h"
#include "debug.h"

/*
 * - same way to handle param (type param1, type param2...)
 * - check valid data in DMA VRAM COPY
 * - check valid data in show_tiles
 */

/*************    Variables  *************/
static   ushort  KANGEN_pal_black[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; /* black palette */
volatile ushort  extWIDTH;
volatile ushort  extHEIGHT;


/************* Video_ScreenSize ************
 * Define size of the screen, in tiles
 * This size is used everywhere so
 * you MUST call this func!
 **************************************/
void Video_ScreenSize( ushort width, ushort height)
{
	extWIDTH = width;
	extHEIGHT = height;	
}


/************* Video_DoDMA_copy ************
 * Use the DMA to copy the map data to  *
 * the GFX RAM.                         *
 * source = array of ulong              *
 * dest   = first number of the tiles   *
 *         (32bytes*existing_tiles)     *
 * len    = length of ulong             *
 *         (32bytes*nbr_tiles)          *
 *                                      *
 *   Made by P.Lee                      *
 ****************************************/
void Video_DoDMA_copy(uint source,ushort dest,ushort len)
{
    register volatile ushort *pw;
    register uint *pl;

    Z80_Pause( );

    /* Point to GFX register */
    pw = (ushort *) GFXCNTL;

    /* Setup DMA length */
    len >>= 1;
    *pw = 0x9300L + (len & 0xff);
    *pw = 0x9400L + ((len >> 8) & 0xff);

    /* Setup DMA address */
    source >>= 1;
    *pw = 0x9500L + (source & 0xff);

    source >>= 8;
    *pw = 0x9600L + (source & 0xff);

    source >>= 8;
    *pw = 0x9700L + (source & 0xff);

    /* Enable DMA */
    pl = (uint *) GFXCNTL;
    *pl = GFX_DMA_ADDR((uint) dest);

    Z80_Release( );
}

/**************** Video_Tile *******************
 * Show a by writing the tile property    		*
 *     priority flipV           tile number     *
 *            |  |  _____________|              *
 * TileDef =  0000 0000 0000 0000               *
 *             ||  |                            *
 *        palette  flipH                        *
 *                                              *
 * fieldAdr = adress of plan of dest            *
 * pri = priority                               *
 *                                              *
 *      Made by Kaneda							*
 *              assume you define the WIDTH     *
 ************************************************/
void Video_Tile(ushort posx, ushort posy, ushort tileDef, ushort fieldAdr)
{
	volatile uint *pl;
	volatile ushort *pw;
	register uint i;

	/* Point to the coordinate we want */
	i = fieldAdr + (  2 * (posx + (extWIDTH*posy) ) ); 
	
	pl = (uint *) GFXCNTL;
	*pl = GFX_WRITE_ADDR(i);

	pw = (ushort *) GFXDATA;
	*pw = tileDef;
}

/**************** Video_Fill ********************
 * Show the tiles by writing the tile           *
 * number (1 to...) to the GFX memory.          *
 *     priority flipV           tile number     *
 *            |  |  _____________|              *
 * TileDef =  0000 0000 0000 0000               *
 *             ||  |                            *
 *        palette  flipH                        *
 *                                              *
 * fieldAdr = adress of plan of dest            *
 * pri = priority                               *
 *                                              *
 *      Made by P.Lee                           *
 *      Modified by Kaneda to allow arg         *
 *              allow Flipping                  *
 *              assume you define the WIDTH     *
 ************************************************/
void Video_Fill(ushort from, ushort to, uchar pal, ushort posx, ushort posy,\
	 ushort fieldAdr, uchar pri, uchar flipV, uchar flipH)
{
	volatile uint *pl;
	volatile ushort *pw;
	register  uint i;
	register uint TileDef;

	/* Point to the coordinate we want (WIDTH=Plane width defined in .H) */
	i = fieldAdr + (  2 * (posx + (extWIDTH*posy) ) ); 
	
	/* We make TileDef (see up)*/
	TileDef = pri&0x01;
 	TileDef <<= 2;
	TileDef |= pal&0x03;
	TileDef <<= 1;
	TileDef |= flipV&0x01;
	TileDef <<= 1;
	TileDef |= flipH&0x01;
	TileDef <<= 11;

	pl = (uint *) GFXCNTL;
	*pl = GFX_WRITE_ADDR(i);

	pw = (ushort *) GFXDATA;
	for (i=from;i < to+1;i++)
		*pw = (TileDef | i);
}




/***************** Video_SetPal *****************
 * Fill the colors data of the given pal		    *
 *                                              *
 *   Made by P.Lee                              *
 *   Modified by Kaneda to use arg and any pal  *
 ************************************************/
void Video_SetPal(uchar palNumber,ushort *palData)
{
	register uint *pl;
  register volatile ushort *pw;
  ushort decal;
  uchar x;

	/* Pal 1 is 32bytes, 16colors of 2bytes each, after Pal 0 */
  decal = palNumber * 32;

	/* Point on adress memory of the pal we want to modif/create */
	pl = (uint *) GFXCNTL;
	*pl = GFX_COLOR_WRITE_ADDR(decal);

	pw = (ushort *) GFXDATA;
	for (x = 0; x < 16; x++)
		*pw=palData[x];
}

/********************** Video_FadeIn ********************
 * Fade in the palette PALNUM to the new palette PALDEST*
 *                                                      *
 * Made by P.Lee                                        *
 * Mod. by Kaneda                                       *
 *   fade to something else than the black pal          *
 ********************************************************/
void Video_FadeIn(palDest, palNum)
ushort *palDest;
uchar palNum;
{
    ushort tmp_pal[16];
    register uint *pl;
    register ushort *pw;
    register ushort *pal_ptr;
    register ushort i, j, c;
    register uint Decal;

    /* see set_colors() */
    Decal = (32 * palNum);

    /* Initialize a temporary palette */
    for (i = 0; i < 16; i++)
        tmp_pal[i] = 0;

    /* Fade in loop */
    for (i = 0; i < 16; i++)
    {
        /* Pointer to palette we want */
        pal_ptr = palDest;
    
        /* Setup GFX to save colors */
        pl = (uint *) GFXCNTL;
        *pl = GFX_COLOR_WRITE_ADDR(Decal);

        /* Setup GFX data register */
        pw = (ushort *) GFXDATA;
        
        /* Increment one color B,G,R at a time */
        for (j = 0; j < 16; j++)
        {
            /* increment BLUE till reaching target color */
            c = tmp_pal[j] & 0xf00;
            if (c < (*pal_ptr & 0xf00))
                tmp_pal[j] += 0x100;
            /* check GREEN */
            c = tmp_pal[j] & 0xf0;
            if (c < (*pal_ptr & 0xf0))
                tmp_pal[j] += 0x10;
            /* check RED */
            c = tmp_pal[j] & 0xf;
            if (c < (*pal_ptr++ & 0xf))
                tmp_pal[j] += 0x1;

            /* Store color */
            *pw = tmp_pal[j];
        }

        /* Wait a sec... */
        for (j = 0; j < 4; j++)
            wait_sync();
    }
}




/*********************** Video_VScroll **********************
 * Scroll the mapfield A                                *
 * pos < 0 , scroll down                                *
 * pos > 0 , scroll up                                  *
 *                                                      *
 * Made by P.Lee                                        *
 * Modified by Kaneda                                   *
 ********************************************************/
void Video_VScroll(posA, posB)
short int posA;
short int posB;
{
    register volatile short int *pw;
    register uint *pl;

    /* Setup GFX pointers */
    pl = (uint *) GFXCNTL;
    *pl = GFX_VERT_SCROLL(0); /* 0 if all scroll else, the number of col */

    pw = (short int *) GFXDATA;
    *pw = posA;
    *pw = posB;
}




/****************** Video_HScroll ***********************
 * Scroll mapfield A and B                              *
 * pos < 0 , scroll left                                *
 * pos > 0 , scroll right                               *
 *                                                      *
 * Made by Kaneda                                       *
 * 0.1 Delete the wait_sync()				*
 ********************************************************/
void Video_HScroll(posA, posB)
short int posA;
short int posB;
{
    register volatile short int *pw;
    register ulong *pl;

    pl = (ulong *) GFXCNTL;
    *pl = GFX_WRITE_ADDR(HSCRL);

    pw = (short int *) GFXDATA;
    *pw = posA;
    *pw = posB;
    /* you must scroll fieldA before fieldB....(?) */
}


/****************** Video_Init ***********************
 * Init/Reset all Video related stuff                *
 *                                                   *
 * Made by Kaneda                                    *
 *****************************************************/
void Video_Init()
{
	uchar i,j;
	
	Video_SetPal(PAL0, KANGEN_pal_black);
  Video_SetPal(PAL1, KANGEN_pal_black);
  Video_SetPal(PAL2, KANGEN_pal_black);
  Video_SetPal(PAL3, KANGEN_pal_black);

  Video_HScroll(0,0);
  Video_VScroll(0,0);
   
	for (j=0;j<extHEIGHT;j++)
	{
		for( i=0; i<extWIDTH; i++)
		{
			Video_Tile( i, j, 0, BPLAN);
			Video_Tile( i, j, 0, APLAN);
		}
	}
}
